import javafx.beans.binding.DoubleBinding;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.EventHandler;
import javafx.geometry.Bounds;
import javafx.scene.control.SkinBase;
import javafx.scene.input.MouseEvent;
import javafx.scene.shape.Arc;
import javafx.scene.shape.Circle;
import javafx.scene.shape.StrokeType;

public class KnobSkin extends SkinBase<Knob> {
    private Circle body;
    private Circle marker;
    private Arc highlight;
    private Arc shadow;
    private Circle minTick;
    private Circle maxTick;

    public KnobSkin(Knob knob) {
        super(knob);

        initChildren();
        initBehavior();
    }

    private void initBehavior() {
        body.radiusProperty().bind(getSkinnable().bodyRadiusProperty());
        body.fillProperty().bind(getSkinnable().bodyColorProperty());
        
        marker.radiusProperty().bind(getSkinnable().markerRadiusProperty());
        marker.fillProperty().bind(getSkinnable().markerColorProperty());
        
        minTick.radiusProperty().bind(getSkinnable().tickRadiusProperty());
        minTick.fillProperty().bind(getSkinnable().tickColorProperty());
        maxTick.radiusProperty().bind(getSkinnable().tickRadiusProperty());
        maxTick.fillProperty().bind(getSkinnable().tickColorProperty());
        
        DoubleBinding borderRadius = getSkinnable().bodyRadiusProperty().add(
                getSkinnable().borderWidthProperty());
        highlight.radiusXProperty().bind(borderRadius);
        highlight.radiusYProperty().bind(borderRadius);
        highlight.strokeProperty().bind(getSkinnable().borderColorProperty());
        highlight.fillProperty().bind(getSkinnable().highlightColorProperty());
        highlight.startAngleProperty().bind(getSkinnable().borderAngleProperty());
        shadow.radiusXProperty().bind(borderRadius);
        shadow.radiusYProperty().bind(borderRadius);
        shadow.strokeProperty().bind(getSkinnable().borderColorProperty());
        shadow.fillProperty().bind(getSkinnable().shadowColorProperty());
        shadow.startAngleProperty().bind(getSkinnable().borderAngleProperty().add(180));

        ChangeListener<Number> listener = new ChangeListener<Number>() {
            @Override
            public void changed(ObservableValue<? extends Number> observable,
                                Number oldValue, Number newValue) {
                getSkinnable().requestLayout();
            }
        };
        getSkinnable().valueProperty().addListener(listener);
        getSkinnable().markerSpacingProperty().addListener(listener);
        getSkinnable().tickSpacingProperty().addListener(listener);

        getSkinnable().setOnMouseDragged(new EventHandler<MouseEvent>() {
            @Override
            public void handle(MouseEvent event) {
                Bounds contentBounds = getSkinnable().getLayoutBounds();
                double cx = contentBounds.getMinX() + contentBounds.getWidth() / 2;
                double cy = contentBounds.getMinY() + contentBounds.getHeight() / 2;
                
                double pointerX = event.getX() - cx;
                double pointerY = event.getY() - cy;
                double pointerAngle = Math.toDegrees(Math.atan2(pointerY, pointerX));
                
                double angle = pointerAngle - getSkinnable().getStartAngle();
                if (pointerY < 0 || pointerX > 0) {
                    angle += 360;
                }
                
                double value = angle / getSkinnable().getTotalAngle() * (
                        getSkinnable().getMaxValue() - getSkinnable().getMinValue());
                value = Math.max(Math.min(value, getSkinnable().getMaxValue()),
                        getSkinnable().getMinValue());
                getSkinnable().setValue(value);
            }
        });
    }

    private void initChildren() {
        initBody();
        initMarker();
        initBorder();
        initTicks();

        getChildren().addAll(highlight, shadow, body, marker, minTick, maxTick);
    }

    private void initTicks() {
        minTick = new Circle();
        maxTick = new Circle();
    }

    private void initBorder() {
        highlight = new Arc();
        highlight.setLength(180);
        highlight.setStrokeType(StrokeType.INSIDE);

        shadow = new Arc();
        shadow.setLength(180);
        shadow.setStrokeType(StrokeType.INSIDE);
    }

    private void initMarker() {
        marker = new Circle();
    }

    private void initBody() {
        body = new Circle();
    }

    @Override
    protected void layoutChildren(double contentX, double contentY,
                                  double contentWidth, double contentHeight) {
        double cx = contentX + contentWidth / 2;
        double cy = contentY + contentHeight / 2;

        layoutBody(cx, cy);
        layoutMarker(cx, cy);
        layoutBorder(cx, cy);
        layoutTicks(cx, cy);
    }

    private void layoutBorder(double cx, double cy) {
        double borderRadius = getSkinnable().getBodyRadius() + getSkinnable().getBorderWidth();
        double borderTheta = Math.toRadians(getSkinnable().getBorderAngle());
        
        highlight.relocate(cx - borderRadius, cy - borderRadius);

        double shadowX = Math.cos(borderTheta) * borderRadius;
        double shadowY = Math.sin(borderTheta) * borderRadius;
        shadow.relocate(cx - shadowX, cy - shadowY);
    }

    private void layoutMarker(double cx, double cy) {
        double markerOrbitRadius = body.getRadius() - marker.getRadius() -
                getSkinnable().getMarkerSpacing();
        double markerAnglePerValue = getSkinnable().getTotalAngle() / (
                getSkinnable().getMaxValue() - getSkinnable().getMinValue());
        double markerAngle = getSkinnable().getStartAngle() +
                getSkinnable().getValue() * markerAnglePerValue;
        double markerTheta = Math.toRadians(markerAngle);
        double markerX = Math.cos(markerTheta) * markerOrbitRadius;
        double markerY = Math.sin(markerTheta) * markerOrbitRadius;

        marker.relocate(cx - marker.getRadius() + markerX, cy - marker.getRadius() + markerY);
    }

    private void layoutBody(double cx, double cy) {
        body.relocate(cx - body.getRadius(), cy - body.getRadius());
    }

    private void layoutTicks(double cx, double cy) {
        double tickOrbitRadius = getSkinnable().getBodyRadius() +
                getSkinnable().getTickRadius() + getSkinnable().getTickSpacing();
        
        double minTickTheta = Math.toRadians(getSkinnable().getStartAngle());
        double minTickX = Math.cos(minTickTheta) * tickOrbitRadius;
        double minTickY = Math.sin(minTickTheta) * tickOrbitRadius;
        minTick.relocate(cx - getSkinnable().getTickRadius() + minTickX,
                cy - getSkinnable().getTickRadius() + minTickY);

        double maxTickTheta = Math.toRadians(getSkinnable().getStartAngle() +
                getSkinnable().getTotalAngle());
        double maxTickX = Math.cos(maxTickTheta) * tickOrbitRadius;
        double maxTickY = Math.sin(maxTickTheta) * tickOrbitRadius;
        maxTick.relocate(cx - getSkinnable().getTickRadius() + maxTickX,
                cy - getSkinnable().getTickRadius() + maxTickY);
    }

    @Override
    protected double computeMinWidth(double height, double topInset,
                                     double rightInset, double bottomInset,
                                     double leftInset) {
        return computePrefWidth(height, topInset, rightInset, bottomInset, leftInset);
    }

    @Override
    protected double computeMaxWidth(double height, double topInset,
                                     double rightInset, double bottomInset,
                                     double leftInset) {
        return computePrefWidth(height, topInset, rightInset, bottomInset, leftInset);
    }

    @Override
    protected double computePrefWidth(double height, double topInset,
                                      double rightInset, double bottomInset,
                                      double leftInset) {
        return rightInset + getSkinnable().getTickRadius() * 2 * 2 + getSkinnable().getTickSpacing() * 2 +
               getSkinnable().getBodyRadius() * 2 + leftInset;
    }

    @Override
    protected double computeMinHeight(double height, double topInset,
                                      double rightInset, double bottomInset,
                                      double leftInset) {
        return computePrefHeight(height, topInset, rightInset, bottomInset, leftInset);
    }

    @Override
    protected double computeMaxHeight(double height, double topInset,
                                      double rightInset, double bottomInset,
                                      double leftInset) {
        return computePrefHeight(height, topInset, rightInset, bottomInset, leftInset);
    }

    @Override
    protected double computePrefHeight(double width, double topInset,
                                       double rightInset, double bottomInset,
                                       double leftInset) {
        return topInset + getSkinnable().getTickRadius() * 2 * 2 + getSkinnable().getTickSpacing() * 2
                + getSkinnable().getBodyRadius() * 2 + bottomInset;
    }
}
